""" Manage the grade CSV file"""

import pandas

from .student_repos import convert_github_url_format
from .utils import error


def parse_and_check(grades_csv_path, csv_cols):
    """Parse the grades CSV file and check that column names are valid"""
    try:
        grades_df = pandas.read_csv(grades_csv_path)
    except pandas.errors.EmptyDataError:
        error(
            "Exception: pandas.errors.EmptyDataError.  Is your", grades_csv_path.name, "file empty?"
        )
    check_csv_column_names(grades_df, csv_cols)
    return grades_df


def check_csv_column_names(df, expected_grade_col_names):
    """This function checks that the provided CSV file has the correct number of coulumns,
    and that each column name matches the expected values for the lab being graded"""

    required_columns = ["Last Name", "First Name", "Net ID"]
    required_columns += expected_grade_col_names

    for required_column in required_columns:
        if required_column not in df.columns:
            error(
                "Grades CSV must contain column '" + required_column + "'.",
                "Current columns:",
                list(df.columns),
            )


def filter_need_grade(df, expected_grade_col_names):
    """Filter down to only those students that need a grade"""
    filtered_df = df[df[df.columns.intersection(expected_grade_col_names)].isnull().any(1)]
    return filtered_df


def match_to_github_url(df_needs_grade, github_csv_path, github_csv_col_name, use_https):
    """Match students to their github URL"""
    try:
        df_github = pandas.read_csv(github_csv_path, index_col=False)
    except pandas.errors.EmptyDataError:
        error(
            "Exception pandas.errors.EmptyDataError. Is your", github_csv_path.name, "file empty?"
        )

    # Strip whitespace from CSV header names
    df_github.rename(columns=lambda x: x.strip(), inplace=True)

    # Rename appropriate column to github url
    df_github.rename(columns={github_csv_col_name: "github_url"}, inplace=True)

    # Convert github URLs to https or SSH
    df_github["github_url"] = df_github["github_url"].apply(
        lambda url: convert_github_url_format(url, use_https)
    )

    # Filter down to relevant columns
    df_github = df_github[["Net ID", "github_url"]]

    # Merge with student dataframe (inner merge will drop students without github URL)
    df_joined = df_needs_grade.merge(df_github)

    return df_joined


def add_group_column_from_csv(df, column_name, groups_csv_path, groups_csv_col_name):
    """Read the group names from the group CSV and join them to the original grades CSV"""
    df_groups = pandas.read_csv(groups_csv_path)

    if column_name in df.columns:
        error(
            "The",
            "'" + column_name + "'",
            "column is used for your groups, but this column already exists in your grade CSV file.",
            "The same column name cannot exist in both places.",
        )

    # Rename appropriate column to group
    df_groups.rename(columns={groups_csv_col_name: column_name}, inplace=True)

    # Filter down to relevant columns
    if "Net ID" not in df_groups.columns:
        error("Your group CSV", "(" + str(groups_csv_path) + ")", "is missing a 'Net ID' column.")
    df_groups = df_groups[["Net ID", column_name]]

    # Merge with student dataframe (inner merge will drop students not in group CSV)
    df_joined = df.merge(df_groups)

    return df_joined


def find_idx_for_netid(df, netid):
    """Find the row index for a given student netid"""
    matches = df.index[df["Net ID"] == netid].tolist()
    if len(matches) != 1:
        error("Could not find netid =", netid, "(find_idx_for_netid)")
    return matches[0]


def get_net_ids(row):
    """Get net IDs from row in grade CSV"""
    return row["Net ID"]


def get_first_names(row):
    """Return first names of group members in the row"""
    return row["First Name"]


def get_last_names(row):
    """Return last names of group members in the row"""
    return row["Last Name"]


def get_concated_names(row):
    """Return a concatenated list of group member names for the row"""
    return ", ".join(
        [
            (first + " " + last + " (" + net_id + ")")
            for (first, last, net_id) in zip(
                get_first_names(row), get_last_names(row), get_net_ids(row)
            )
        ]
    )



""" Module to manage each item that is to be graded"""


import csv
import json
import shutil
import sys

import pandas

from .utils import CallbackFailed, TermColors, print_color, error
from . import grades_csv, utils


class GradeItem:
    """Class to track each item that needs to be graded (ie, each item for which a grading callback
    function will be invoked.  This may be used to grade one or more columns from the CSV file."""

    def __init__(
        self, grader, csv_col_names, fcn, max_points, feedback_filename, feedback_col_name, help_msg
    ) -> None:
        self.grader = grader
        self.csv_col_names = csv_col_names
        self.fcn = fcn
        self.max_points = max_points
        self.feedback_filename = feedback_filename
        self.feedback_col_name = feedback_col_name
        self.feedback_enabled = self.feedback_filename or self.feedback_col_name
        self.help_msg = help_msg

        # Feedback comments
        self.feedback_list_path = self.grader.work_path / (str(csv_col_names) + ".json")
        self.feedback_list = []
        if self.feedback_list_path.is_file():
            with open(self.feedback_list_path, encoding="utf-8") as f:
                self.feedback_list = json.load(f)

        # Feeback file directory
        if self.feedback_filename:
            self.feedback_dir_path = (
                grader.grades_csv_path.parent / "feedback" / self.feedback_filename
            )
            self.feedback_zip_path = (
                grader.grades_csv_path.parent / "feedback" / (self.feedback_filename + ".zip")
            )
            self.feedback_dir_path.mkdir(exist_ok=True, parents=True)

    def run_grading(self, student_grades_df, row, callback_args):
        """Run the grading process for this item"""
        net_ids = grades_csv.get_net_ids(row)
        first_names = grades_csv.get_first_names(row)
        last_names = grades_csv.get_last_names(row)
        num_group_members = len(net_ids)
        concated_names = grades_csv.get_concated_names(row)
        num_group_members_need_grade_per_col = self.num_grades_needed(row)

        # variable to flag if build needs to be performed
        # initialize to True as the code must be built at least once
        # (will be false if user chooses to just re-run and not re-build)
        build = True

        if sum(self.num_grades_needed(row)) == 0:
            # No one in the group needs grades for this
            print_color(
                TermColors.BLUE,
                "Grade already exists for ",
                self.csv_col_names,
                "(skipping)",
            )
            return

        while True:
            print_color(
                TermColors.BLUE,
                "Running callback function",
                "(" + str(self.fcn.__name__) + ")",
                "to grade",
                str(self.csv_col_names) + ":",
            )

            scores = None

            # Build it and run
            try:
                scores = self.fcn(
                    **callback_args,
                    csv_col_names=self.csv_col_names,
                    points=self.max_points,
                    build=build and not self.grader.run_only,
                )
            except CallbackFailed as e:
                print_color(TermColors.RED, repr(e))
                break
            except KeyboardInterrupt:
                print("")
            else:
                print_color(TermColors.BLUE, "Callback returned:", scores)

            # reset the flag
            build = True

            # If we are only building the code in preparation of grading later,
            # then exit now before asking for a grade
            if self.grader.build_only:
                break

            for i, col in enumerate(self.csv_col_names):
                if num_group_members_need_grade_per_col[i] < num_group_members:
                    print_color(
                        TermColors.YELLOW,
                        "Warning:",
                        num_group_members - num_group_members_need_grade_per_col[i],
                        "group member(s) already have a grade for",
                        col,
                        "; this grade will be overwritten.",
                    )

            if scores is None:
                # If no score was returned by the callback function, prompt the user for a score.
                try:
                    scores, feedback = self._get_scores(concated_names)
                except KeyboardInterrupt:
                    print_color(TermColors.RED, "\nExiting")
                    sys.exit(0)
            else:
                # If score(s) were returned, make sure the length matches the number of columns to be graded
                scores = utils.ensure_tuple(scores)
                expected_lenth = len(self.csv_col_names)
                if self.feedback_enabled:
                    expected_lenth += 1
                if len(scores) != expected_lenth:
                    error(
                        "The callback should be grading",
                        len(self.csv_col_names),
                        "column(s)",
                        "(" + str(self.csv_col_names) + "),",
                        "but",
                        len(scores),
                        "values were returned.",
                        "Since feedback is enabled, you should return one extra item that is the feedback, which can be an empty string for no feedback."
                        if self.feedback_enabled
                        else "",
                    )
                if self.feedback_enabled:
                    feedback = scores[-1]
                    scores = scores[:-1]

            if scores == "s":
                break
            if scores == "b":
                continue
            if scores == "r":
                # run again, but don't build
                build = False
                continue

            # Record score
            for (first_name, last_name, net_id) in zip(first_names, last_names, net_ids):
                row_idx = grades_csv.find_idx_for_netid(student_grades_df, net_id)

                for (i, col) in enumerate(self.csv_col_names):
                    student_grades_df.at[row_idx, col] = scores[i]

                if self.feedback_col_name:
                    existing_feedback = student_grades_df.at[
                        row_idx, self.feedback_col_name
                    ].strip()
                    if existing_feedback and (existing_feedback[-1] != "."):
                        existing_feedback += ". "

                    # Append new feedback
                    student_grades_df.at[row_idx, self.feedback_col_name] = (
                        existing_feedback + feedback
                    )

                # Save feedback to a file
                if self.feedback_filename:
                    feedback_file_path = self.feedback_dir_path / (
                        first_name
                        + "_"
                        + last_name
                        + "_"
                        + net_id
                        + "_feedback-"
                        + self.feedback_filename
                        + ".txt"
                    )
                    with open(feedback_file_path, "a", encoding="utf-8") as f:
                        f.write(feedback + "\n")

                    # Create zip archive
                    if self.feedback_zip_path.is_file():
                        self.feedback_zip_path.unlink()
                    shutil.make_archive(
                        self.feedback_zip_path.with_suffix(""), "zip", self.feedback_dir_path
                    )

            student_grades_df.to_csv(
                str(self.grader.grades_csv_path),
                index=False,
                quoting=csv.QUOTE_ALL,
            )
            break

    def num_grades_needed(self, row):
        """Return the number of total grades needed across all group members for
        each grade column in this row."""
        empty_per_col = []
        for col in self.csv_col_names:
            empty_cnt = 0
            for grade in row[col]:
                if pandas.isnull(grade):
                    empty_cnt += 1
            empty_per_col.append(empty_cnt)
        return empty_per_col

    def _get_scores(self, names):
        """Prompts the user for a score for the grade column(s)."""
        fpad = " " * 8
        fpad2 = " " * 4
        pad = 10
        feedback = ""
        scores = []

        for (i, grade_col) in enumerate(self.csv_col_names):
            points = self.max_points[i] if self.max_points else None
            while True:
                print("")
                if self.help_msg:
                    print_color(TermColors.BOLD, self.help_msg[i])

                ################### Build input menu #######################
                input_txt = (
                    TermColors.BLUE
                    + "Enter a grade for "
                    + names
                    + ", "
                    + (TermColors.UNDERLINE + grade_col + TermColors.END + TermColors.BLUE)
                    + ":\n"
                )

                # Add current feedback
                if self.feedback_enabled:
                    input_txt += (
                        fpad
                        + "Pending feedback: "
                        + TermColors.END
                        + feedback
                        + TermColors.BLUE
                        + "\n"
                    )

                # Add score input
                input_txt += (
                    fpad2
                    + (("0-" + str(points)) if points else "#").ljust(pad)
                    + "Enter a score to finish and save\n"
                )

                # Enter feedback
                allowed_feedback = {}
                if self.feedback_enabled:
                    input_txt += (
                        fpad2
                        + "str".ljust(pad)
                        + "Enter a string with any new feedback, or select from previous feedback:\n"
                    )
                    for idx, f in enumerate(self.feedback_list):
                        input_txt += fpad2 + ("f" + str(idx)).ljust(pad + 2) + f + "\n"
                        allowed_feedback["f" + str(idx)] = f

                    input_txt += fpad2 + "'c'".ljust(pad) + "Clear entered feedback\n"
                    allowed_feedback["c"] = ""

                input_txt += fpad2 + "'s'".ljust(pad) + "Skip to next student\n"
                allowed_cmds = ["s"]

                if self.grader.allow_rebuild:
                    input_txt += fpad2 + "'b'".ljust(pad) + "Build and run again\n"
                    allowed_cmds.append("b")
                if self.grader.allow_rerun:
                    input_txt += fpad2 + "'r'".ljust(pad) + "Re-run"
                    if self.grader.allow_rebuild:
                        input_txt += " (w/o rebuild)"
                    input_txt += "\n"
                    allowed_cmds.append("r")

                # Remove trailing ", " and terminate
                input_txt += ">>> " + TermColors.END

                ################### Get and handle user input #######################
                txt = input(input_txt)

                # Check for commands
                if txt in allowed_cmds:
                    scores = txt
                    break

                # Check for integer input
                try:
                    score = float(txt)
                    if (points is None) or (0 <= score <= points):
                        scores.append(score)
                        break
                    print("Invalid input. Try again.")
                except ValueError:
                    pass

                # Check for feedback input
                if txt in allowed_feedback:
                    if txt == "c":
                        feedback = ""
                        continue
                    feedback_to_add = allowed_feedback[txt]

                else:
                    txt = txt.capitalize()
                    if txt not in self.feedback_list:
                        self.feedback_list.append(txt)
                        with open(self.feedback_list_path, "w", encoding="utf-8") as f:
                            json.dump(self.feedback_list, f)
                    feedback_to_add = txt

                # Assume input is feedback
                if feedback and feedback[-1] != ".":
                    feedback += ". "
                feedback += feedback_to_add

            # If non-integer returned, then user asked for something like re-run, so stop prompting for grades and exit this function
            if isinstance(scores, str):
                return (scores, "")

        return (scores, feedback)



""" Main ygrader module"""
from collections import defaultdict
import pathlib
import enum
import re
import zipfile
import time
import os
import shutil
from typing import Callable
import inspect
import pandas


from . import grades_csv
from . import utils, student_repos
from .grading_item import GradeItem
from .utils import CallbackFailed, directory_is_empty, print_color, TermColors, error, warning


class CodeSource(enum.Enum):
    """Used to indicate whether the student code is submitted via LearningSuite or Github"""

    LEARNING_SUITE = 1
    GITHUB = 2


class Grader:
    """Grader class"""

    def __init__(
        self,
        lab_name: str,
        grades_csv_path: pathlib.Path,
        work_path: pathlib.Path = pathlib.Path.cwd() / "temp",
    ):
        """
        Parameters
        ----------
        lab_name: str
            Name of the lab/assignment that you are grading (ie. 'lab3').
            This is used for folder naming, logging mesasge, etc., and passed back to your callback functions.
        grades_csv_path: pathlib.Path
            Path to CSV file with student grades exported from LearningSuite.  You need to export netid, first
            and last name, and any grade columns you want to populate.
        work_path: pathlib.Path
            Path to directory where student files will be placed.  For example, if you pass in '.', then student
            code would be placed in './lab3'.  By default the working path is a "temp" folder created in your working directory.
        """
        self.lab_name = lab_name
        self.grades_csv_path = pathlib.Path(grades_csv_path).resolve()

        # Make sure grades csv exists, and that file is writable
        if not self.grades_csv_path.is_file():
            error("grades_csv_path", "(" + str(grades_csv_path) + ")", "does not exist")
        try:
            with open(grades_csv_path, "a", encoding="utf-8"):
                pass
        except PermissionError:
            error(
                "You do not have permissions to modify the grades_csv_path file",
                "(" + str(grades_csv_path) + ").",
                "Is this file open and locked?",
            )

        # Create a working directory
        self.work_path = pathlib.Path(work_path).resolve()
        self.work_path = self.work_path / lab_name

        # Read CSV and make sure it isn't empty
        try:
            pandas.read_csv(self.grades_csv_path)
        except pandas.errors.EmptyDataError:
            error("Your grades csv", "(" + str(grades_csv_path) + ")", "appears to be empty")

        # Initialize other class members
        self.items = []
        self.code_source = None
        self.prep_fcn = None
        self.learning_suite_submissions_zip_path = None
        self.github_csv_path = None
        self.github_csv_col_name = None
        self.github_tag = None
        self.github_https = None
        self.groups_csv_path = None
        self.groups_csv_col_name = None
        self.set_other_options()

    def add_item_to_grade(
        self,
        csv_col_names,
        grading_fcn,
        max_points=None,
        feedback_filename=None,
        feedback_col_name=None,
        help_msg=None,
    ):
        """Add a new item you want to grade.

        Parameters
        ----------
        csv_col_names: (str | list of str)
            The column name(s) from your grading CSV file that you want to grade.
        grading_fcn: Callable
            The callback function that will perform all your grading work. Your callback function will be provided with the following arguments:

              * lab_name (*str*): This will pass back the lab name you passed to *__init__*.
                Useful if you use the same callback function to grade multiple different assignments.

              * student_code_path (*pathlib.Path*): The location where the unzipped/cloned student files are stored.
              * cols_to_grade (*str*): The current CSV column being graded. Typically only needed if you are grading
                multiple different items.
              * points (*int*): The maximum number of points possible for the item being graded, used for validating the
                grade when prompting the user to input a grade.  If your callback function automatically calcuates and
                returns a grade, this argument is ignored.
              * first_names: (*list(str)*) First name(s) of students in the group
              * last_names: (*list(str)*) Last name(s) of students in the group
              * net_ids: (*list(str)*) Net ID(s) of students in the group.
              * build: (*bool*) Whether files should be built/compiled.
              * run: (*bool*) Whether milestone should be run.

            In addition, if your grades CSV exported from Learning Suite has the following information, it will also be
            provided to your callback function:

              * section: (*str*) Student section number, assuming 'Section Number' was contained in grades_csv exported
                from Learning Suite.
              * homework_id: (*str*) Student homework ID, assuming 'Course Homework ID' was contained in grades_csv
                exported from Learning Suite.

            Your callback should return *None* or an *int*/*float* (or a list of *int*/*float* if you are grading multiple
            columns in this item).  If you return *None*, then the user will be prompted to input a grade.  If you already
            know the grade you want to assign, and don't want to prompt the user, return the grade value(s).

            If there's a problem with the student's submission and you want to skip them, then `raise CallbackFailed`.
            You can provide an argument to this exception with any error message you want printed.

            Since your callback functions will be provided with many arguments, it's best to use keyword arguments:

            .. highlight:: python
            .. code-block:: python

                def my_callback(**kw):
                    lab_name = kw["lab_name"]
                    first_name = kw["first_names"][0]
        max_points: int | list of int
            (Optional) Number of max points for the graded column(s).
        feedback_filename: str
            (Optional) Filename for feedback that will be zipped and can be uploaded to LearningSuite.
            This feedback zip file will be stored in a 'feedback' directory, located with your grades CSV.
        feedback_col_name: str
            (Optional) Alterantively, feedback can be saved to a column in the grade CSV file.  To do this, provide the name of CSV column.
        help_msg: str | list of str
            (Optional) When the script asks the user for a grade, it will print this message first.  This can be a helpful
            reminder to the TAs of a grading rubric, things they should watch out for, etc.  If you are grading multiple
            CSV columns, then you should provide a separate help message per column.
        """
        # Check data types
        if not isinstance(grading_fcn, Callable):
            error("'grading_fcn' must be a callable function")

        # Make these into lists, even if there is only one item
        csv_col_names = utils.ensure_tuple(csv_col_names)
        if max_points:
            max_points = utils.ensure_tuple(max_points)

            # Check that # of CSV columns to grade matches # of points list
            if len(csv_col_names) != len(max_points):
                error(
                    "'csv_col_names'",
                    csv_col_names,
                    "has list length =",
                    len(csv_col_names),
                    "but 'max_points'",
                    max_points,
                    "has list length =",
                    str(len(max_points)) + ".",
                    "They must be equal.",
                )

        if help_msg:
            help_msg = utils.ensure_tuple(help_msg)

            # Check that # of CSV columns to grade matches # of help_msg list
            if len(csv_col_names) != len(help_msg):
                error(
                    "'csv_col_names'",
                    csv_col_names,
                    "has list length =",
                    len(csv_col_names),
                    "but 'help_msg'",
                    help_msg,
                    "has list length =",
                    str(len(help_msg)) + ".",
                    "They must be equal.",
                )

        df = pandas.read_csv(self.grades_csv_path)
        for col_name in csv_col_names:
            if col_name not in df:
                error(
                    "Provided grade column name",
                    "(" + col_name + ")",
                    "does not exist in grades_csv_path",
                    "(" + str(self.grades_csv_path) + ").",
                    "Columns:",
                    list(df.columns),
                )

        if feedback_filename is not None and feedback_col_name is not None:
            error("Provide only one of feedback_filename or feedback_col_name")
        if feedback_col_name and feedback_col_name not in df:
            error(
                "Provided feedback_col_name",
                "(" + feedback_col_name + ")",
                "does not exist in grades_csv_path",
                "(" + str(self.grades_csv_path) + ").",
                "Columns:",
                list(df.columns),
            )

        item = GradeItem(
            self,
            csv_col_names,
            grading_fcn,
            max_points,
            feedback_filename,
            feedback_col_name,
            help_msg,
        )
        _verify_callback_fcn(grading_fcn, item)
        self.items.append(item)

    def set_submission_system_learning_suite(self, zip_path):
        """
        Call this function if you are using student submissions from Learning Suite.

        Parameters
        ----------
        zip_path: pathlib.Path | str
            Path to the zip file that was downloaded from Learning Suite using *Batch Download*.
        """
        zip_path = pathlib.Path(zip_path).resolve()

        self.code_source = CodeSource.LEARNING_SUITE
        self.learning_suite_submissions_zip_path = zip_path

        if not zip_path.is_file():
            error("Provided zip_path", zip_path, "does not exist")

    def set_submission_system_github(
        self, tag, github_url_csv_path, repo_col_name="github_url", use_https=False
    ):
        """
        Call this function if you are using student submissions on Github.

        Parameters
        ----------
        tag: str
            The tag on the students github repository that is used for the submission.
        github_url_csv_path: pathlib.Path | str
            The path to a CSV file containing student Github repository URLs.  This CSV file must have a
            column called 'Net ID'. The repos can be listed as an http:// address or in SSH format (git@).
            They will be coverted as needed depending on the *use_https* argument.
        repo_col_name: str
            The column name in the CSV file that contains the Github URLs.
        use_https: bool
            By default SSH will be used to clone the student repos.  If you want to use an access token or stored
            credentials over https, set this to True.
        """
        self.code_source = CodeSource.GITHUB
        self.github_csv_path = pathlib.Path(github_url_csv_path).resolve()
        self.github_csv_col_name = repo_col_name
        self.github_tag = tag
        self.github_https = use_https

        if not github_url_csv_path.is_file():
            error(
                "Provided github_url_csv_path",
                "(" + str(github_url_csv_path) + ")",
                "does not exist",
            )

        df = pandas.read_csv(github_url_csv_path)
        if repo_col_name not in df:
            error(
                "Provided repo_col_name",
                "(" + repo_col_name + ")",
                "is not a column in github_url_csv_path",
                "(" + str(github_url_csv_path) + ")",
            )

    def set_learning_suite_groups(self, csv_path, col_name="group"):
        """
        This function is used to provide treams to the grader when grading a team-based assignment from
        Learning Suite.  (If grading from Github, the Github URL will be used to determine teams)

        Parameters
        ----------
        csv_path: str
            The path to a CSV file containing group names. This CSV file must have a column called 'Net ID'.
        col_name: str
            The column name in the CSV file that indicates the group name.
        """
        if self.code_source != CodeSource.LEARNING_SUITE:
            error(
                "Please call set_submission_system_learning_suite() before calling set_learning_suite_groups()."
            )

        self.groups_csv_path = pathlib.Path(csv_path).resolve()
        self.groups_csv_col_name = col_name

        if not csv_path.is_file():
            error("Provided groups csv_path", csv_path, "does not exist")

        df = pandas.read_csv(self.groups_csv_path)
        if col_name not in df:
            error("Provided groups col_name", col_name, "does not exist in", csv_path)

    def set_other_options(
        self,
        format_code=False,
        build_only=False,
        run_only=False,
        allow_rebuild=False,
        allow_rerun=True,
        prep_fcn=None,
    ):
        """
        This can be used to set other options for the grader.

        Parameters
        ----------
        format_code: bool
            Whether you want the student code automatically formatted using clang-format
        build_only: bool
            Whether you only want to build and not run/grade the students code.  This will be passed to your
            callback function, and is useful for labs that take a while to build.  You can build all the code in
            one pass, then return and grade the code later.
        run_only: bool
            Whether you only want to run/grade and not build the students code.  This will be passed to your
            callback function, and is useful for labs that take a while to build.  You can build all the code
            in one pass, then return and grade the code later.
        allow_rebuild: bool
            By default, the program will pass build=True and run=True to your callback on the first invocation,
            and then allow the grader the option to "re-run" the student's code, where build=False and run=True
            would be provided.  If you want to allow the grader to
            specifically request a rebuild, set this to True (default False).
        allow_rerun: bool
            By default, the program will pass build=True and run=True to your callback on the first invocation,
            and then allow the grader the option to "re-run" the student's code, where build=False and run=True
            would be provided.  If your grader can't rerun without rebuilding, then set this to False (default True).
        prep_fcn: Callable
            If you are grading multiple items, then you can use this optional callback to do any one-time prep work.
            This callback is provided the same arguments as the grading callback function, except for *cols_to_grade*
            and *max_points*.  You should not return any value from this callback, but you can `raise CallbackFailed` to skip the student.
        """
        self.format_code = format_code
        self.build_only = build_only
        self.run_only = run_only
        self.allow_rebuild = allow_rebuild
        self.allow_rerun = allow_rerun
        if prep_fcn and not isinstance(prep_fcn, Callable):
            error("The 'prep_fcn' argument must provide a callable function pointer")
        self.prep_fcn = prep_fcn
        if prep_fcn:
            _verify_callback_fcn(prep_fcn, item=None)

        if not (self.allow_rebuild or self.allow_rerun):
            error("At least one of allow_rebuild and allow_rerun needs to be True.")

    def _validate_config(self):
        """Check that everything has been configured before running"""
        # Check that callback function has been set up
        if not self.items:
            error(
                "Before calling run(), you must call add_item_to_grade() to add an item to grade."
            )

        # Check that submission source is set
        if self.code_source is None:
            error(
                "Before calling run(), you must set a submission source by either calling "
                + "set_submission_system_learning_suite() or set_submission_system_github()."
            )

    def _get_all_csv_cols_to_grade(self):
        """Collect all columns that will be graded into a single list"""
        return [col for item in self.items for col in item.csv_col_names]

    def run(self):
        """Call this to start (or resume) the grading process"""

        self._validate_config()

        # Print starting message
        print_color(TermColors.BLUE, "Running grader for", self.lab_name)

        # Read in CSV and validate.  Print # students who need a grade
        student_grades_df = grades_csv.parse_and_check(
            self.grades_csv_path, self._get_all_csv_cols_to_grade()
        )

        # Convert columns
        for item in self.items:
            if item.feedback_col_name:
                student_grades_df[item.feedback_col_name] = student_grades_df[
                    item.feedback_col_name
                ].fillna("")

        # Filter by students who need a grade
        grades_needed_df = grades_csv.filter_need_grade(
            student_grades_df, self._get_all_csv_cols_to_grade()
        )
        print_color(
            TermColors.BLUE,
            str(
                grades_csv.filter_need_grade(
                    grades_needed_df, self._get_all_csv_cols_to_grade()
                ).shape[0]
            ),
            "students need to be graded.",
        )

        # Add column for group name to DataFrame.
        # For github, students are grouped by their Github repo URL.
        # For learning suite, if set_groups() was never called, then  students are placed in groups by Net ID (so every student in their own group)
        grouped_df = self._group_students(student_grades_df)

        # Create working path directory
        self._create_work_path()

        # For learning suite, unzip their submission and add a column that points to it
        if self.code_source == CodeSource.LEARNING_SUITE:
            pass
            # self._unzip_submissions()
            # sys.exit(0)
            # grouped_df = self._add_submitted_zip_path_column(grouped_df)

        self._run_grading(student_grades_df, grouped_df)

    def _run_grading(self, student_grades_df, grouped_df):
        # Loop through all of the students/groups and perform grading
        for _, row in grouped_df.iterrows():
            first_names = grades_csv.get_first_names(row)
            last_names = grades_csv.get_last_names(row)
            net_ids = grades_csv.get_net_ids(row)
            concated_names = grades_csv.get_concated_names(row)

            # Check if student/group needs grading
            num_group_members_need_grade_per_item = [
                item.num_grades_needed(row) for item in self.items
            ]

            if sum(sum(s) for s in num_group_members_need_grade_per_item) == 0:
                # This student/group is already fully graded
                continue

            # Print name(s) of who we are grading
            student_work_path = self.work_path / utils.names_to_dir(
                first_names, last_names, net_ids
            )
            print_color(
                TermColors.PURPLE,
                "\nGrading: ",
                concated_names,
                "-",
                student_work_path.relative_to(self.work_path.parent),
            )

            # Get student code from zip or github.  If this fails it returns False.
            # Code from zip will return modified time (epoch, float). Code from github will return True.
            success = self._get_student_code(row, student_work_path)
            if not success:
                continue

            # Format student code
            if self.format_code:
                print_color(TermColors.BLUE, "Formatting code")
                utils.clang_format_code(student_work_path)

            callback_args = {}
            callback_args["lab_name"] = self.lab_name
            callback_args["student_code_path"] = student_work_path
            callback_args["run"] = not self.build_only
            callback_args["first_names"] = first_names
            callback_args["last_names"] = last_names
            callback_args["net_ids"] = net_ids
            if "Section Number" in row:
                callback_args["section"] = row["Section Number"]
            if "Course Homework ID" in row:
                callback_args["homework_id"] = row["Course Homework ID"]

            if self.prep_fcn is not None:
                try:
                    self.prep_fcn(
                        **callback_args,
                        build=not self.run_only,
                    )
                except CallbackFailed as e:
                    print_color(TermColors.RED, repr(e))
                    continue
                except KeyboardInterrupt:
                    pass

            # Loop through all items that are to be graded
            for item in self.items:
                item.run_grading(student_grades_df, row, callback_args)

    def _unzip_submissions(self):
        with zipfile.ZipFile(self.learning_suite_submissions_zip_path, "r") as f:
            for zip_info in f.infolist():
                # Remove old zip file if it exists
                unpack_path = self.work_path / zip_info.filename
                if unpack_path.is_file():
                    unpack_path.unlink()

                # Unzip
                f.extract(zip_info, self.work_path)

                # Fix timestamp
                date_time = time.mktime(zip_info.date_time + (0, 0, -1))
                os.utime(unpack_path, (date_time, date_time))

    def _add_submitted_zip_path_column(self, df):
        # Map dataframe index to student zip file
        df_idx_to_zip_path = {}

        for index, row in df.iterrows():
            net_ids = grades_csv.get_net_ids(row)

            # Find all submissions that belong to the group
            zip_matches = []
            for net_id in net_ids:
                zip_matches.extend(list(self.work_path.glob("*_" + net_id + "_*.zip")))
            if len(zip_matches) == 0:
                # print("No zip files match", group_name)
                continue
            if len(zip_matches) > 1:
                # Multiple submissions -- get the latest one
                zip_matches.sort(key=lambda x: x.stat().st_mtime, reverse=True)

            df_idx_to_zip_path[index] = zip_matches[0]

        df["submitted_zip_path"] = pandas.Series(df_idx_to_zip_path)
        df["submitted_zip_path"] = df["submitted_zip_path"].fillna(value="")
        return df

    def _group_students(self, df):
        if self.code_source == CodeSource.GITHUB:
            # For Github source, group name is simply github URL
            df = grades_csv.match_to_github_url(
                df, self.github_csv_path, self.github_csv_col_name, self.github_https
            )

            df_needs_grades = grades_csv.filter_need_grade(df, self._get_all_csv_cols_to_grade())

            print_color(
                TermColors.BLUE,
                str(df_needs_grades.shape[0]),
                "of these students have a github URL.",
            )
            groupby_column = "github_url"

        elif self.groups_csv_path is None:
            # No groups, so just group by netid

            groupby_column = "group_id"
            df = df.copy()
            df[groupby_column] = df["Net ID"]
        else:
            # Add group column from given CSV
            groupby_column = "group_id"
            df = grades_csv.add_group_column_from_csv(
                df, groupby_column, self.groups_csv_path, self.groups_csv_col_name
            )

            # Check how many students remain
            df_needs_grades = grades_csv.filter_need_grade(df, self._get_all_csv_cols_to_grade())
            print_color(
                TermColors.BLUE,
                str(df_needs_grades.shape[0]),
                "of these students belong to a group.",
            )

        # Group students into their groups
        return df.groupby(groupby_column).agg(list).reset_index()

    def _get_student_code(self, row, student_work_path):
        if self.code_source == CodeSource.GITHUB:
            return self._get_student_code_github(row, student_work_path)

        # else:
        return self._get_student_code_learning_suite(row, student_work_path)

    def _get_student_code_github(self, row, student_work_path):
        student_work_path.mkdir(parents=True, exist_ok=True)

        # Clone student repo
        print("Student repo url: " + row["github_url"])
        if not student_repos.clone_repo(row["github_url"], self.github_tag, student_work_path):
            return False
        return True

    def _get_student_code_learning_suite(self, row, student_work_path):
        print("Extracting submitted files for", grades_csv.get_concated_names(row), "...")
        if student_work_path.is_dir() and not directory_is_empty(student_work_path):

            # Code already extracted from Zip, return
            print("  Files already extracted previously.")
            return True

        student_work_path.mkdir(parents=True, exist_ok=True)

        # Keep track of last file extracted by name
        extracted_by_name = {}

        # Track how many files of each name are extracted so we can warn about duplicate submissions
        count_by_filename = defaultdict(int)

        with zipfile.ZipFile(self.learning_suite_submissions_zip_path, "r") as top_zip:

            # Loop through all files in top-level zip file
            for file in top_zip.infolist():

                if file.is_dir():
                    continue

                # Loop through everyone in the group
                for netid in grades_csv.get_net_ids(row):

                    # Check if file belongs to student
                    match = re.match("^.*?_" + netid + "_(.*)$", file.filename)
                    if not match:
                        continue

                    # Handle regular files (not zip files)
                    if not file.filename.lower().endswith(".zip"):
                        extract_to_name = match.group(1)

                        count_by_filename[extract_to_name] += 1

                        # If we've already extracted a file of this name, don't overwrite if older
                        if extract_to_name in extracted_by_name:
                            if file.date_time <= extracted_by_name[extract_to_name].date_time:
                                continue
                        top_zip.extract(file, student_work_path)
                        extracted_by_name[extract_to_name] = file

                        # Rename to remove student name/netid from file
                        unpack_old_path = student_work_path / file.filename
                        unpack_new_path = student_work_path / extract_to_name
                        unpack_old_path.rename(unpack_new_path)

                        # Restore timestamp
                        date_time = time.mktime(file.date_time + (0, 0, -1))
                        os.utime(unpack_new_path, (date_time, date_time))
                        continue

                    # Otherwise this is a zip within zip. Open it up and collect contained files
                    with zipfile.ZipFile(top_zip.open(file)) as inner_zip:
                        for file2 in inner_zip.infolist():
                            if file2.is_dir():
                                continue
                            count_by_filename[file2.filename] += 1

                            # If we've already extracted a file of this name, don't overwrite if older
                            if file2.filename in extracted_by_name:
                                if file2.date_time <= extracted_by_name[file2.filename].date_time:
                                    continue

                            inner_zip.extract(file2, student_work_path)
                            extracted_by_name[file2.filename] = file2

                            # Restore timestamp
                            unpack_path = student_work_path / file2.filename
                            date_time = time.mktime(file2.date_time + (0, 0, -1))
                            os.utime(unpack_path, (date_time, date_time))

        # Print what was extracted
        for k in sorted(extracted_by_name.keys()):
            print("  ", k, end=" ")
            if count_by_filename[k] > 1:
                print_color(
                    TermColors.YELLOW,
                    "(" + str(count_by_filename[k]),
                    "versions submitted, using last modified.)",
                )
            else:
                print()

        # Return success if at least one file is obtained
        if len(extracted_by_name) == 0:
            print_color(TermColors.YELLOW, "No submission")
            return False

        return True

    def _create_work_path(self):
        if self.code_source == CodeSource.LEARNING_SUITE:
            if self.work_path.is_dir() and (
                self.learning_suite_submissions_zip_path.stat().st_mtime
                > self.work_path.stat().st_mtime
            ):
                shutil.rmtree(self.work_path)

        if not self.work_path.is_dir():
            print_color(TermColors.BLUE, "Creating", self.work_path)
            self.work_path.mkdir(exist_ok=True, parents=True)


def _verify_callback_fcn(fcn, item):
    callback_args = [
        "lab_name",
        "student_code_path",
        "run",
        "build",
        "first_names",
        "last_names",
        "net_ids",
    ]
    if item:
        if item.max_points:
            callback_args.append("max_points")

        # If this is a fcn for a graded item (not a prep-only function), then
        # this argument is required.
        callback_args.append("csv_col_names")

    callback_args_optional = [
        "section",
        "homework_id",
    ]

    # Check that callback function(s) are valid
    argspec = inspect.getfullargspec(fcn)

    # Check that kwargs is enabled
    if argspec.varkw is None:
        error(
            "Your callback function",
            "(" + fcn.__name__ + ")",
            "should accept keyward arguments (**kw). This is needed because the grader may provide "
            + "different optional arguments to your callback depending on what data it has available",
            "(" + ",".join(callback_args_optional) + ").",
        )

    # Check that all named arguments are valid
    for i, named_arg in enumerate(argspec.args):
        # Skip special arguments
        if named_arg in ("self", "cls") and i == 0:
            continue
        if (named_arg not in callback_args) and (named_arg not in callback_args_optional):
            error(
                "Your callback function",
                "(" + fcn.__name__ + ")",
                "takes a named argument",
                "'" + named_arg + "'",
                "but this is not provided by the grader. Please remove this argument or the grader "
                + "will not be able to call your callback function correctly. Available callback arguments:",
                str(callback_args),
            )
        elif named_arg not in callback_args:
            warning(
                "Your callback function",
                "(" + fcn.__name__ + ")",
                "takes a named argument",
                "'" + named_arg + "'",
                "but this argument is not always provided by the grader.",
                "If it is missing from your grades CSV file this will cause a runtime error.",
                "Please consider use keyword arguments (**kw) instead.",
            )



from setuptools import setup

setup(
    name="ygrader",
    packages=["ygrader"],
    version="1.1.14",
    description="Grading scripts used in BYU's Electrical and Computer Engineering Department",
    author="Jeff Goeders",
    author_email="jeff.goeders@gmail.com",
    license="MIT",
    url="https://github.com/byu-cpe/ygrader",
    install_requires=["pandas>=1.0.0"],
)
